#include <parallel/parallel.h>
#include <sys/tmp_ctl.h>
#include <usclkc.h>
#include <signal.h>

typedef unsigned long us_ticks;

/* number of nodes participating in the test */
shared int num_nodes = 0;

/* each processor has its own virtual pid */
private int vpid;
int affinity[32];

extern int exit(); /* external declaration for the C library routine */

/* quit -- terminate self and all others in this process group */
void quit()
{
	killpg(getpgrp(0),SIGQUIT);
}

/*	main -- run test on each of the spinlocks
 *		ARGUMENTS
 *		-i<num_procs> run a single test of each synch primitive on num_procs
 *			processors
 *		-p<num_procs> run a single test of each synch primitive on 
 *			2..num_procs processors
 *		DEFAULTS
 *		run all of the tests with on each of 2..NUMPROCS processors
 *		(where NUMPROCS 2 less than the number of cpus online)
 */
main(argc,argv)
char **argv;
int argc;
{
	int index = 0;
	int i;
	int noaffinity = 0;

	while (--argc > 0) {
		if(**++argv == '-')
			switch((*argv)[1]) {
			case 'i':
				index = atoi(&(*argv)[2]);
				break;
			case 'p':
				num_nodes = atoi(&(*argv)[2]);
				break;
			default:
				printf("unknown argument %s\n",*argv);
			}
	}

	usclk_init(); 				/* initialize usec clock */

	signal(SIGQUIT,exit); 		/* terminate cleanly upon SIGQUIT */
	signal(SIGINT,quit);  		/* send SIGQUIT to terminate my process group */

	if (index > 0) num_nodes = index;
	else if (num_nodes == 0) {
		num_nodes = cpus_online();	/* find out how many processors */
		num_nodes -= 2; 			/* leave two nodes free for the system */
	}

	simple_barrier_init(num_nodes);

	/* create (num_nodes - 1) children (num_nodes processes total) */
	vpid = 0; /* assign vpid 0 to parent */
	for(i=1;i<num_nodes;i++) {
		int fval = fork();
		if (!fval) {
			vpid = i; /* each child gets unique vpid */
			break;
		}
		if (fval == -1) {
			printf("unable to fork virtual process %d\n", i);
			exit(1);
		}
	}

	if ((affinity[vpid] = tmp_affinity(vpid)) == -1)
		printf("    unable to set affinity for vpid %d\n",vpid);

	simple_barrier();

	if (vpid == 0) {
		printf("%s",BARRIERSTRING);

		for(i=0;i<num_nodes;i++) if (affinity[i] == -1) noaffinity = 1;
		if(noaffinity) printf("-no"); 
		else printf("-");

		printf("affinity\n");
	}

	BARRIER/**/_barrier_init();
	simple_barrier();
	BARRIER/**/_barrier();

	{
#define ITERATIONS 100000
		us_ticks start_time, end_time; 
		start_time = GETUSCLK(); 

		for(i=ITERATIONS;i>=0;i--)  {
			BARRIER/**/_barrier();
#ifdef DEBUGGING
			if (vpid == 0 && ((i % 100) == 0)) printf("%d done\n",i);
#endif DEBUGGING
		}

		end_time = GETUSCLK(); 
		if (vpid == 0) 
		printf("Barrier test %d processors, %d iterations, time = %d usec, each = %d usec\n",
			num_nodes, ITERATIONS, end_time - start_time, 
			(end_time - start_time)/ITERATIONS);
	}
}
