#include <parallel/parallel.h>
#include <sys/tmp_ctl.h>
#include <usclkc.h>
#include <signal.h>
#include "processor.h"
#include "synchtest.h"
#include "anderson.h"

typedef struct SI {
	struct SI *next;
	int locked;
	int dummy[2]; /* pad to  fill out cache line */
} spinitem;


/* global structures (one for each type of spin lock)  */
shared slock_t simplelock;
shared spinitem **mcslock;
shared Lock *andersonlock;

/* structure for collecting timing results for each of the processors */
shared long elapsed_time[NUMPROCS];

/* number of nodes participating in the test */
shared int num_nodes = 0;

/* each processor has its own virtual pid */
private int vpid;

extern int exit(); /* external declaration for the C library routine */

/* quit -- terminate self and all others in this process group */
void quit()
{
	killpg(getpgrp(0),SIGQUIT);
}

/*	main -- run test on each of the spinlocks
 *		ARGUMENTS
 *		-i<num_procs> run a single test of each synch primitive on num_procs
 *			processors
 *		-p<num_procs> run a single test of each synch primitive on 
 *			1..num_procs processors
 *		DEFAULTS
 *		run all of the tests with on each of 1..NUMPROCS processors
 *		(where NUMPROCS is the 2 less than the number of cpus online)
 */
main(argc,argv)
char **argv;
int argc;
{
	int index = 0;
	int i;

	while (--argc > 0) {
		if(**++argv == '-')
			switch((*argv)[1]) {
			case 'i':
				index = atoi(&(*argv)[2]);
				break;
			case 'p':
				num_nodes = atoi(&(*argv)[2]);
				break;
			default:
				printf("unknown argument %s\n",*argv);
			}
	}

	usclk_init(); 				/* initialize usec clock */

	signal(SIGQUIT,exit); 		/* terminate cleanly upon SIGQUIT */
	signal(SIGINT,quit);  		/* send SIGQUIT to terminate my process group */

	if (index > 0) num_nodes = index;
	else if (num_nodes == 0) {
		num_nodes = cpus_online();	/* find out how many processors */
		num_nodes -= 2; 			/* leave two nodes free for the system */
	}

	barrier_init(num_nodes);

	/* create (num_nodes - 1) children (num_nodes processes total) */
	vpid = 0; /* assign vpid 0 to parent */
	for(i=1;i<num_nodes;i++) {
		int fval = fork();
		if (!fval) {
			vpid = i; /* each child gets unique vpid */
			break;
		}
		if (fval == -1) {
			printf("unable to fork virtual process %d\n", i);
			exit(1);
		}
	}

	if (tmp_affinity(vpid) == -1)
		printf("unable to set affinity for vpid %d\n",vpid);

	barrier();

	timing_tests(index);
}

delay()
{
	register int i;
	for(i=0;i<3;i++);
}

void null_loop_test()
{
	int j;

	asm("# --null loop test--");
	RUN_TEST_ALL ("null", 0, , { ; });
}

void delay_test()
{
	int j;

	asm("# --delay test--");
	RUN_TEST_ALL ("delay", 0, , { delay();});
}

void simple_lock_test(index)
int index;
{
	int j;

	asm("# --simple test-and-test-and-set--");
	if (vpid == 0) {
		s_init_lock(&simplelock); /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("simple", (index - 1), 
			register slock_t *lptr = &simplelock, 
			{simple_lock(lptr); simple_unlock(lptr);});
	} else {
		for(j = 0; j < num_nodes; j++) {
			RUN_TEST_ALL ("simple", j, 
				register slock_t *lptr = &simplelock, 
				{simple_lock(lptr); simple_unlock(lptr);});
		}
	}
}

void simple_lock_exp_test(index)
int index;
{
	int j;

	asm("# --test-and-set w/ exp backoff--");
	if (vpid == 0) {
		s_init_lock(&simplelock); /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("simpleexp", (index - 1), 
			register slock_t *lptr = &simplelock, 
			{simpleexp_lock(lptr); simpleexp_unlock(lptr);});
	} else {
		for(j = 0; j < num_nodes; j++) {
			RUN_TEST_ALL ("simpleexp", j, 
				register slock_t *lptr = &simplelock, 
				{simpleexp_lock(lptr); simpleexp_unlock(lptr);});
		}
	}
}

void simple_lock_exp_test_delay(index)
int index;
{
	int j;

	asm("# --test-and-set w/ exp backoff (delay) --");
	if (vpid == 0) {
		s_init_lock(&simplelock); /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("simpleexp-delay", (index - 1), 
			register slock_t *lptr = &simplelock, 
			{simpleexp_lock(lptr); delay(); simpleexp_unlock(lptr);});
	} else {
		for(j = 0; j < num_nodes; j++) {
			RUN_TEST_ALL ("simpleexp-delay", j, 
				register slock_t *lptr = &simplelock, 
				{simpleexp_lock(lptr); delay(); simpleexp_unlock(lptr);});
		}
	}
}

void simple_lock_test_delay(index)
int index;
{
	int j;

	asm("# --simple test-and-test-and-set (delay)--");
	if (vpid == 0) {
		s_init_lock(&simplelock); /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("simple-delay", (index - 1), 
			register slock_t *lptr = &simplelock, 
			{simple_lock(lptr); delay(); simple_unlock(lptr);});
	} else {
		for(j = 0; j < num_nodes; j++) {
			RUN_TEST_ALL ("simple-delay", j, 
				register slock_t *lptr = &simplelock, 
				{simple_lock(lptr); delay(); simple_unlock(lptr);});
		}
	}
}

void mcs_lock_test(index)
int index;
{
	spinitem *myitem;
	int j;

	myitem = (spinitem *) shmalloc(sizeof(spinitem)); /* allocate my link */

	asm("# --mcs spin-lock--");
	if (vpid == 0) {
		mcslock = 0; /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("mcs", (index - 1), 
			register spinitem ***lptr = &mcslock; register spinitem *myspin = myitem,
			{mcs_lock(lptr,myspin); mcs_unlock(lptr,myspin);});
	} else {
		for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("mcs", j,
			register spinitem ***lptr = &mcslock; register spinitem *myspin = myitem,
			{mcs_lock(lptr,myspin); mcs_unlock(lptr,myspin);});
		}
	}
}

private int *myl;

void anderson_lock_backoff_test(index)
int index;
{
	int j;

	asm("# --anderson backoff spin-lock --");
	if (vpid == 0) {
		andersonlock = (Lock *) shmalloc(sizeof(Lock));
		ANDERSON_LOCK_INIT(andersonlock);
	}

	if (index > 0) {
		anderson_init_b(vpid,1);
		RUN_TEST_ALL ("anderson-backoff", (index - 1), 
			register Lock *lptr = andersonlock , 
			{register LockList* list = anderson_lock_b(lptr);
				anderson_unlock_b(list);});
	} else {
		for(j = 0; j < num_nodes; j++) {
		anderson_init_b(vpid,1);
		RUN_TEST_ALL ("anderson-backoff", j,
			register Lock *lptr = andersonlock ,
			{register LockList* list = anderson_lock_b(lptr);
				anderson_unlock_b(list);});
		}
	}
}

void anderson_lock_backoff_test_delay(index)
int index;
{
	int j;

	asm("# --anderson backoff spin-lock/delay--");
	if (vpid == 0) {
		andersonlock = (Lock *) shmalloc(sizeof(Lock));
		ANDERSON_LOCK_INIT(andersonlock);
	}

	if (index > 0) {
		anderson_init_b(vpid,1);
		RUN_TEST_ALL ("anderson-backoff-delay", (index - 1), 
			register Lock *lptr = andersonlock , 
			{register LockList* list = anderson_lock_b(lptr);
				delay();
				anderson_unlock_b(list);});
	} else {
		for(j = 0; j < num_nodes; j++) {
		anderson_init_b(vpid,1);
		RUN_TEST_ALL ("anderson-backoff-delay", j,
			register Lock *lptr = andersonlock ,
			{register LockList* list = anderson_lock_b(lptr);
				delay();
				anderson_unlock_b(list);});
		}
	}
}

void mcs_lock_test_delay(index)
int index;
{
	spinitem *myitem;
	int j;

	myitem = (spinitem *) shmalloc(sizeof(spinitem)); 

	asm("# --mcs spin-lock/delay--");
	if (vpid == 0) {
		mcslock = 0; /* initialize shared lock */
	}

	if (index > 0) {
		RUN_TEST_ALL ("mcs-delay", (index - 1), 
			register spinitem ***lptr = &mcslock; register spinitem *myspin = myitem,
			{mcs_lock(lptr,myspin); delay(); mcs_unlock(lptr,myspin);});
	} else {
		for(j = 0; j < num_nodes; j++) {
		RUN_TEST_ALL ("mcs-delay", j, 
			register spinitem ***lptr = &mcslock; register spinitem *myspin = myitem,
			{mcs_lock(lptr,myspin); delay(); mcs_unlock(lptr,myspin);});
		}
	}
}

timing_tests(index)
int index;
{
	null_loop_test(); 
	delay_test(); 
	simple_lock_test(index);
	simple_lock_test_delay(index);
	simple_lock_exp_test(index);
	simple_lock_exp_test_delay(index);
	anderson_lock_backoff_test(index);
	anderson_lock_backoff_test_delay(index);
	mcs_lock_test(index);
	mcs_lock_test_delay(index);
}
